require 'olympic/bracket/single_elimination'

module Olympic
  module Bracket
    # A basic bracket.  This defines the basic API that all brackets
    # should have.  In order for a bracket to be compatible with
    # Olympic, only these things need to be implemented.
    class Base

      # Initialize the bracket.
      #
      # @param tournament [Olympic::Tournament]
      # @param teams [Array<Olympic::Team>]
      def initialize(tournament, teams)
        @tournament = tournament
        @teams = teams
      end

      # @!method call
      #   Initializes the tournament to handle the bracket.  Sets up
      #   all matches and sources in accordance to the bracket
      #   itself.
      #
      #   @return [void]
      #
      # @!method matches
      #   Returns the number of expected matches in the bracket.  If
      #   the number of matches is variable, it will return an array
      #   of numbers, which contains potential values for the number
      #   of matches in the bracket.
      #
      #   @return [Numeric, Array<Numeric>]
      #
      # @!method rounds
      #   Returns the number of expected rounds in the bracket.  If
      #   the number of rounds is variable, it will return an array
      #   of numbers, which contains potential values for the number
      #   of rounds in the bracket.
      #
      #   @return [Numeric, Array<Numeric>]
      #
      # @!method rounded_teams
      #   In an elimination tournament, the number of teams has to be
      #   rounded to the nearest power of two to calculate the number
      #   of byes.  If this is not necessary, it will return the
      #   number of teams.
      #
      #   @return [Numeric]
      #
      # @!method byes
      #   The number of byes that are required in the bracket.  This
      #   should be greater than or equal to zero.
      #
      #   @return [Numeric]
      #
      # @!method first_round_matches
      #   The number of matches in the first round.  This will always
      #   be a finite number, so this method will never return an
      #   array of numbers.
      #
      #   @return [Numeric]
      %(call matches rounds rounded_teams byes
        first_round_matches).map(&:to_sym).each do |name|
        define_method(name) do
          _not_implemented_error(name)
        end
      end

      private

      # Raises an error stating that the given method wasn't
      # implemented.  This is the default action the API methods on
      # this class takes; a subclass should redefine the methods to
      # perform an action.
      #
      # @note This should not be used outside of the base class.  If
      #   A `NotImplementedError` is to be raised outside of the base
      #   class, it should be raised with `raise`.
      # @raise [NotImplementError] on call.
      # @return [void]
      def _not_implemented_error(name)
        raise NotImplementedError, "`#{name}` is not implemented " \
          "on `#{self.class}`"
      end
    end
  end
end
