require 'memoist'

module Olympic
  module Rating
    class Glicko < Base
      class Formula

        extend Memoist

        # The default rating.  Glicko states to use 1500 as the
        # default.
        DEFAULT_RATING = 1500

        # The default RD.  Glicko states to use 350 as the default.
        DEFAULT_RATING_DERIVATION = 350

        # This is the "certainty decay," or the constant value that
        # is used in Step 1 of the Glicko rating system.  It denotes
        # the decay of certainty of a player's rating over a given
        # time (see {Glicko} for a definition of time).  The default
        # chosen here expects a typical RD of 50, with 10 time units
        # before a decay to 350.  The formula would be (in LaTeX):
        #
        #   c = \sqrt{\frac{350^2-50^2}{10}}
        #
        # However, in order to optimize operations, this is c^2.  So
        # the actual value of this would be:
        #
        #   c = \frac{350^2-50^2}{10}
        CERTAINTY_DECAY = 12_000

        Q  = Math.log2(10).fdiv(400)
        Q2 = Q ** 2

        def initialize(unit, matches)
          @unit = unit
          @matches = matches
        end

        def call
          flush_cache
          @unit.rating = new_rating
          @unit.derivation = new_derivation
        end

        # This returns the current rating derivation of the unit.  It
        # clamps the rating derivation to between 30 and 350.  The
        # Glicko system states that it must be less than or equal to
        # 350, but later recommends not letting the rating derivation
        # drop below 30.
        #
        # @return [Numeric]
        def derivation
          if @unit.unrated?
            350
          elsif @unit.time_passed == 0
            @unit.derivation
          else
            a = [
              30,
              Math.sqrt((@unit.derivation) ** 2 + CERTAINTY_DECAY * @unit.time_passed),
              350
            ].sort[1]
          end
        end
        memoize :derivation

        private

        def new_rating
          @unit.rating +
            Q.fdiv(1.0.fdiv(derivation ** 2) + 1.0.fdiv(delta)) * summation
        end
        memoize :new_rating

        def new_derivation
          Math.sqrt(1.0.fdiv(
            1.0.fdiv(derivation ** 2) + 1.0.fdiv(delta)))
        end
        memoize :new_derivation

        def summation
          @matches.inject(0) do |memo, value|
            e = epsilon(value[:rating], value[:derivation])
            g = gamma(value[:derivation])
            memo + g * (value[:outcome] - e)
          end
        end
        memoize :summation

        def delta
          1.0.fdiv(Q2 * delta_summation)
        end
        memoize :delta

        def delta_summation
          @matches.inject(0) do |memo, value|
            e = epsilon(value[:rating], value[:derivation])
            g = gamma(value[:derivation])
            memo + (g ** 2) * e * (1 - e)
          end
        end
        memoize :delta_summation

        def gamma(derivation)
          1.0.fdiv Math.sqrt(1 + (3 * Q2 * (derivation ** 2)).fdiv(Math::PI ** 2))
        end
        memoize :gamma

        def epsilon(rating, derivation)
          1.0.fdiv(1 + 10 **
            (gamma(derivation) * (@unit.rating - rating)).fdiv(-400))
        end
        memoize :epsilon

      end
    end
  end
end
